#----------------------------
# RETO PREDICTIVE MODELLING 
# CAJAMAR 
#
# Fecha: 03/2017 
# Alexandra Abos & Xisca Pericas 
#----------------------------

#--------------------------
# Librerias y entorno 
#--------------------------
library(data.table) 
library(ggplot2) ## Gráficos 
library(ggthemr) ## Temas gráficos 
library(zoo) ## Formatos fecha 
library(lubridate) ## Formatos fecha 
library(h2o) ## Libreria h2o 


#--------------------------
# Funciones 
#--------------------------

LoadInH2o <-function(data.to.load, factors=NULL, destination_frame=NULL){
  write.csv(data.to.load, file=paste0(temp.folder, "temp.csv"), row.names=F)
  data.h2o.raw <- h2o.importFile(paste0(temp.folder, "temp.csv"), parse=FALSE)
  data.h2o.setup <- h2o.parseSetup(data.h2o.raw)
  data.h2o.setup$column_types[match(factors, data.h2o.setup$column_names)]<-"Enum"
  if(is.null(destination_frame)){
    destination_frame <- deparse(substitute(data.to.load))
  }
  data.h2o <- h2o.parseRaw(data.h2o.raw, col.types=data.h2o.setup$column_types, destination_frame=destination_frame)
  file.remove(paste0(temp.folder, "/temp.csv"))
  return(data.h2o)
}

# Generar una instancia en h2o con 10g usando los CPU disponibles. 
localH2O <- h2o.init(ip = 'localhost', port = 13462, max_mem_size = '10g', nthreads=-2)

#--------------------------
# DATA  
#--------------------------

# Cargar el train 
data.train <- as.data.table(read.table(paste0(getwd(),'/dades/noves/train2.txt'),header = T, sep = '|', stringsAsFactors = F))
# Quitar variable de énero por motivos éticos  
data.train[,Socio_Demo_04 := NULL]
# Formatear correctamente la fecha 
data.train[,date:=as.Date(as.yearmon(Cod_Fecha))]

# Quitar el último producto comprado (esto será usado como tarrget)
## Last Products 
last.products <- data.train[order(date), .(Cod_Prod =Cod_Prod[.N]), by = ID_Customer]
setkey(last.products,ID_Customer,Cod_Prod)

## Train sin el target 
setkey(data.train, ID_Customer, Cod_Prod)
data.train.without.last <- data.train[!last.products, ]

#******************************************************************************
#                      FEATURE ENGINEERING 
#******************************************************************************

# --------------------------------------------------------------------------------------------------------------------------------------
# VARIABLES DE TIEMPO 
# Generar variables relacionadas con la periodicidad de compra, primera compra, última compra, etc. para cada customer 
# --------------------------------------------------------------------------------------------------------------------------------------

## Añadir la fecha del dia del procesado 
data.train.without.last[,date_today := today()]
data.train.without.last[,diff_date_today := as.numeric(difftime(date_today,date,units = 'days'))]

## Variables de tiempo 
temps_var_per_customer <- data.train.without.last[order(date),.(
  total_productes_comprats = .N,  ## Total productos comprados por customer 
  temps_desde_primera_compra = diff_date_today[1],  ##Tiempo desde la primera compra  hasta hoy 
  temps_desde_darrera_compra = diff_date_today[.N] ## Tiempo desde la última compra hasta hoy 
), by = ID_Customer]

## Lag entre compras (tiempo entre compras )
lag_entre_compres <- data.train.without.last[order(date)][, .(lag_entre_compres = mean(as.numeric(diff.POSIXt(date)))), by = ID_Customer]
## Customers que solo tienen una compra el lag es NaN por tanto se pone como NA (missing value) 
lag_entre_compres[is.nan(lag_entre_compres), lag_entre_compres := NA]
## Merge de datasets 
setkey(lag_entre_compres,ID_Customer)
setkey(temps_var_per_customer, ID_Customer)
temps_var_per_customer <-  temps_var_per_customer[lag_entre_compres,on = 'ID_Customer']

## Añadir al dataset original 
data.train.without.last <- data.train.without.last[temps_var_per_customer, on = 'ID_Customer']

## Añadir número de compras dividido por meses 
compres.per.customer.per.mes <- data.train.without.last[,.N,by = c('ID_Customer','month')]
compres.per.customer.dcast <- dcast.data.table(compres.per.customer.per.mes, ID_Customer ~ month, value.var = 'N')
setnames(compres.per.customer.dcast,grep('[0-9]',colnames(compres.per.customer.dcast),value = T), paste0('month_', grep('[0-9]',colnames(compres.per.customer.dcast),value = T)))
compres.per.customer.dcast[is.na(compres.per.customer.dcast)] <- 0

setkey(temps_var_per_customer,ID_Customer)
setkey(compres.per.customer.dcast,ID_Customer)
temps_var_per_customer <- temps_var_per_customer[compres.per.customer.dcast]

# --------------------------------------------------------------------------------------------------------------------------------------
# MATRIX: HISTORICO DE PRODUCTOS 
# Generar una matriz  que contiene: como columnas cada uno de los Cod_Prod de train/test y como filas el número de productos de cada 
# Cod_Prod comprados por el Customer ID. Una fila por Customer_ID. 
# --------------------------------------------------------------------------------------------------------------------------------------

## Table de customers con un solo producto
customer_one_product <- data.train[,uniqueN(Cod_Prod), by = ID_Customer][V1 == 1, ]
customer_one_product <- data.train[ID_Customer %in% customer_one_product[,ID_Customer], .SD, .SDcols = c('ID_Customer','Cod_Prod',grep('Socio',colnames(data.train),value = T)), ]

## Añadir variables socio (se usaran las variables de la penúltima compra) 
prelast_socio <- data.train.without.last[order(date), .SD[.N], by = ID_Customer, .SDcols = grep('Socio_', colnames(data.train), value = T)]
data.train.without.last[,ind := 1]
pivot_product <- dcast.data.table(data.train.without.last[order(date)],ID_Customer ~ Cod_Prod, fun.aggregate = sum, value.var = 'ind')
pivot_product <- pivot_product[prelast_socio, on = 'ID_Customer']

## Unir tablas 
pivot_product_all <- rbind(pivot_product, customer_one_product[,.SD, .SDcols = setdiff(colnames(customer_one_product),'Cod_Prod')], fill = T)
pivot_product_all[is.na(pivot_product_all)] <- 0

## Adaptar nombres (para no usar sólo números)
setnames(pivot_product_all, grep('^[0-9]', colnames(pivot_product_all), value = T), paste0(grep('^[0-9]', colnames(pivot_product_all), value = T), '_bought_before'))

## Pivotar tabla para obtener una columna por producto y una fila por customer 
customer.per.previous.product <- pivot_product_all[,sapply(.SD, sum), .SDcols = grep('_bought_before', colnames(pivot_product_all), value = T)]
sort(customer.per.previous.product, decreasing = T)/data.train.without.last[,uniqueN(ID_Customer)]*100

# --------------------------------------------------------------------------------------------------------------------------------------
# CLUSTER CUSTOMERS CON LAS VARIABLES 
# Crear un cluster para los distintos customers usando las variables anterirmente procesadas. 
# --------------------------------------------------------------------------------------------------------------------------------------

## Juntar dataset con productos  y dataset con características 
setkey(pivot_product_all, ID_Customer)
setkey(temps_var_per_customer, ID_Customer)
customer.complet <- pivot_product_all[temps_var_per_customer]

## Añadir el penúltimo producto comprado como variable
penultim.products <- data.train.without.last[order(date), .(penultim_cod_prod =Cod_Prod[.N]), by = ID_Customer]
setkey(penultim.products,ID_Customer)
setkey(customer.complet,ID_Customer)
customer.complet <- customer.complet[penultim.products]

## Crear un Cluster usando K-MEANS 
my.factors <- setdiff(colnames(customer.complet),c(grep('month',colnames(customer.complet),value = T),c('Socio_Demo_01','Socio_Demo_02','Socio_Demo_03','total_productes_comprats','temps_desde_primera_compra','temps_desde_darrera_compra','lag_entre_compres')))
customer.complet.h2o <- LoadInH2o(customer.complet, factors = my.factors, destination_frame = 'customer.complet')
# Usamos como predictores todas las variables 
my.predictors <- setdiff(colnames(customer.complet),c('ID_Customer'))
cluster.customer <- h2o.kmeans(customer.complet.h2o, x = my.predictors, nfolds = 5, k = 4, model_id = 'kmean.per.customer.1')

# Prediction 
customer.prediction <- as.data.table(h2o.predict(cluster.customer, customer.complet.h2o))
customer.complet[,cluster := customer.prediction$predict]





# ------------------------------------------
# GRID DE PRODUCTOS A RECOMENDAR
# ------------------------------------------
# Table con todos los productos para cada customer ID 
products_bought <- data.table(expand.grid(last_cod_prod = data.train[,unique(Cod_Prod)], ID_Customer = data.train[,unique(ID_Customer)]))
## Añadir variable de last_product_bought (target) para aquellos productos que si fueron comprados los ultimos 
products_bought_by_customer <- last.products[,.(last_cod_prod,ID_Customer)][,last_product_bought := 1]
setkey(products_bought_by_customer,last_cod_prod,ID_Customer)
# Join
products_bought_by_customer[,last_cod_prod := as.factor(last_cod_prod)]
products_bought[,last_cod_prod := as.factor(last_cod_prod)]
product_customer <- products_bought_by_customer[products_bought, on = c('last_cod_prod','ID_Customer')]
product_customer[is.na(last_product_bought), last_product_bought := 0]

# Unir con las variables de customer 
setkey(product_customer, ID_Customer)
setkey(customer.complet, ID_Customer)
products_bought <- product_customer[customer.complet]


#******************************************************************************
#                     MODELS AND OPTIMIZATION 
#******************************************************************************

## Columnas tipo factor 
my.factors <- setdiff(colnames(customer.complet),c(grep('month',colnames(customer.complet),value = T),c('Socio_Demo_01','Socio_Demo_02','Socio_Demo_03','total_productes_comprats','temps_desde_primera_compra','temps_desde_darrera_compra','lag_entre_compres')))

## Subir los datos a h2o (loadinH2o: escribe los datos en un .csv y luego los importa a h2o y convierte las variables factores a enum) 
# Cluster 0
train.cluster0.h2o <- LoadInH2o(train.set[cluster == 0, ], factors = my.factors, destination_frame = 'data.train.cluster0')
test.cluster0.h2o <- LoadInH2o(test.set[cluster == 0, ], factors = my.factors, destination_frame = 'data.test.cluster0')

# Cluster 1
train.cluster1.h2o <- LoadInH2o(train.set[cluster == 1, ], factors = my.factors, destination_frame = 'data.train.cluster1')
test.cluster1.h2o <- LoadInH2o(test.set[cluster == 1, ], factors = my.factors, destination_frame = 'data.test.cluster1')

# Cluster 2
train.cluster2.h2o <- LoadInH2o(train.set[cluster == 2, ], factors = my.factors, destination_frame = 'data.train.cluster2')
test.cluster2.h2o <- LoadInH2o(test.set[cluster == 2, ], factors = my.factors, destination_frame = 'data.test.cluster2')

# Cluster 3
train.cluster3.h2o <- LoadInH2o(train.set[cluster == 3, ], factors = my.factors, destination_frame = 'data.train.cluster3')
test.cluster3.h2o <- LoadInH2o(test.set[cluster == 3, ], factors = my.factors, destination_frame = 'data.test.cluster3')

## Optimizatcion: se usa un modelo de Random Forest con una optimización de parametros usando un Random Grid Search. 
my.predictors <- setdiff(colnames(train.set),c('ID_Customer','last_product_bought','Socio_Demo_04','id','cluster'))

## Parameters 
# Grid search 
hyper_params = list( ntrees = c(15L,30L,50L,80L,100L), ##Tenemos un total de 117 variables por tanto el tamaño de arboles se limita 
                     nbins = c(5L, 20L, 40L, 60L, 80L),  
                     nbins_cats = c(20L, 50L, 94L), 
                     max_depth = c(1L,5L, 10L,  20L), 
                     min_rows = c(100L, 10000L, 50000L, 335060L), 
                     max_after_balance_size = c(0.5,1,2,5),
                     sample_rate = c(2/3,4/5,1),
                     col_sample_rate_per_tree = c(2/3,4/5,1))

# Search a random subset of these hyper-parmameters. Max runtime 
# and max models are enforced, and the search will stop after we 
# don't improve much over the best 5 random models.
search_criteria = list(strategy = "RandomDiscrete", 
                       max_models = 60, ## Se realizan un total de 60 iteraciones 
                       stopping_metric = "AUC", 
                       stopping_tolerance = 0.00001, 
                       stopping_rounds = 4, 
                       seed = 123456)

## For each cluster 
for (cluster in 0:3) {
  training.frame <- get(paste0('train.cluster',cluster,'.h2o'))
  validation.frame <- get(paste0('test.cluster',cluster,'.h2o'))
  
  ## Grid Search 
  rf_grid <- h2o.grid("randomForest", 
                      grid_id = paste0('cluster_',cluster,'grid'),
                      x = my.predictors, 
                      y = 'last_product_bought', 
                      training_frame = training.frame,
                      validation_frame = validation.frame,
                      binomial_double_trees = T, 
                      balance_classes = T, 
                      seed = 123456,
                      hyper_params = hyper_params,
                      search_criteria = search_criteria)
  
  rf_grid_sorted <- h2o.getGrid(grid_id = paste0('cluster_',cluster,'grid'), sort_by = "auc")
  print(rf_grid_sorted)
  ## Save 
  save(rf_grid_sorted, file = paste0(getwd(),'/rf_cluster_grid/rf_grid_sorted.Rda'))
  
  # Best Model 
  best_model <- h2o.getModel(rf_grid_sorted@model_ids[[60]])
  h2o.saveModel(best_model, paste0(getwd(),'/rf_cluster_grid/',best_model@model_id), force = T)
}

### Prediction of the cluster best models 
cluster_grids <- grep('^cluster_[0-9]grid$',h2o.ls()$key,value = T)

## Se usa el Area Bajo la Curva (AUC) para seleccionar el mejor modelo de cada cluster 
sapply(1:length(cluster_grids),function(cluster_ind) {
  rf_grid_sorted <- h2o.getGrid(cluster_grids[cluster_ind], sort_by = 'auc')
  best_model <- h2o.getModel(rf_grid_sorted@model_ids[[60]])
  ## Hacemos las predicciones de cada dataset para cada cluster
  prediction.table <- as.data.table(h2o.predict(best_model, get(paste0('test.cluster',cluster_ind-1,'.h2o'))))
  test.set[cluster == cluster_ind-1, prediction := prediction.table$p1]
})

#******************************************************************************
#                     VALIDATION 
#******************************************************************************

## Se calcula el Model Score: esta función ordena los productos segun su probabilidad de ser comprados (calculado con los 
## modelos de RF). Entonces se le da a cada producto un valor de orden (del 1 para el que tiene menor probabilidad a 94 para el que
## tiene mayor probabilidad de ser comprado). Finalmente se multiplica el valor del orden con la variable de converger o no (o se compra
## o no). El mejor score que se podria obtener seria 94*el numero total de customers. 

ModelScore <- function(dt, prediction_column, plot.histogram = FALSE) {
  ## Ordenar dataset en funcion de la predicción y dar un valor de orden para todos los productos agrupando por customer
  dt.validate <- dt[order(ID_Customer,get(prediction_column)), .(
    order_prediction = 1:.N, ## from 1 to N (94), 
    last_product_bought, ## Variable dicotómica de 0/1 (no comprado - comprado) 
    get(prediction_column)
  ), by = ID_Customer]
  dt.validate[,model_score := order_prediction * last_product_bought]
  
  ## Calcular el Model score de cada customer. Este model score se calcula como el valor del order prediction (que va de 1 a 94) multiplicado por la variable dicotómica target (0-1)
  ## Entonces, a mayor order_prediction (es decir, a menor valor de probabilidad del producto de ser comprado), menor posición en el ranking y por tanto un mayor model_score (i.e. producto
  ## queda en la posición de recomendación 30 y al final es ese producto el que fue comprado, el model_score será de 30). 
  
  ## Normalización del model score sobre el máximo: 
  model_score_total <- dt.validate[,.(model_score = sum(model_score)), by = ID_Customer][,sum(model_score)] / (94*dt[,uniqueN(ID_Customer)])  ##En nuestro caso, el peor caso seria que el producto comprado fuera el último que se recomendó 
  
  ## Pintar el histograma del model_score si se decea 
  if (isTRUE(plot.histogram)) {
    p1 <- ggplot(dt.validate[,.(model_score = sum(model_score)), by = ID_Customer], aes(x = model_score)) + geom_histogram(bins = 94, fill = '#0EB3F1', color = 'black') + ggtitle('Histogram of Customers vs Model Score')
  }
  
  return(list(model_score_total,p1))
}

ModelScore(test.set,'prediction',plot.histogram = TRUE)


#******************************************************************************
#                     SCORE TEST SET 
#******************************************************************************

# Para el scoring del test se ha usado la misma metodologia: 
# - Cargar datos 
# - Transformar las variables 
# - Cargar en H2o 
# - Computar los clusters 
# - Separar el frame por clusters 
# - Computar las predicciones 
# - Coger los productos con mayor prediccion para cada customer 




